from setuptools import setup, find_packages

setup(
    name='fairpy',
    packages=find_packages(),
    version='0.1.0',
    description='FairPy is a comprehensive Python Library for Machine Learning Fairness',
    url='https://github.com/brandeis-machine-learning/FairPy',
    license='MIT',
    classifiers=[
    "Programming Language :: Python",
    "Programming Language :: Python :: 3.8",
    "Programming Language :: Python :: 3.9",
    ],
)