# @Author  : Peizhao Li <peizhaoli05@gmail.com>
# @License : BSD-2-Clause

from ._adult import Adult
from ._compas import Compas
from ._bank import Bank
from ._credit import Credit
from ._german import German
from ._oulad import Oulad
from ._dutch import Dutch
from ._xing import Xing

__all__ = [
    "Adult",
    "Compas",
    "Bank",
    "Credit",
    "German",
    "Oulad",
    "Dutch",
    "Xing",
]
